﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MegaChip8EmulatorLib
{
    public class MegaChip8
    {
        // Mémoire principale
        // Chip8 possède "4 ko de mémoire 8bit", soit 4096 blocs de 8bit
        // On transpose ca en un tableau de 4096 "byte", le type byte étant un entier non signé codé sur 8 bits
        private byte[] _memory = new byte[4096];
        
        // Program Counter
        // Chip8 possède "un compteur PC, 12bit"
        // Un type de 12bit n'existe pas, on va donc utiliser un type de borne supérieure, en l'occurence ushort (entier non signé sur 16 bits)
        // La documentation dit qu'un programme démarre à l'adresse 0x200h
        private ushort _PC = 0x200;

        // Registres
        // Chip8 possède "16 registres 8bit nommés V0, V2, V3… VF"
        // Pour travailler avec ces 16 registres plus facilement, nous allons en faire un tableau de 16 valeurs (indexé de 0x0 à 0xF)
        private byte[] _V = new byte[16];
        // Plus "1 registre 16bit nommé I"
        private ushort _I;

        // Timers 8bit
        private byte _DT = 0;
        public byte _ST = 0;

        // Charge une ROM en mémoire
        public void Load(byte[] rom)
        {
            // On fait une bête copie en mémoire de la ROM
            // Mais attention, la doc dit que "les programmes Chip8
            // commencent à l'adresse 0x200h, car les 512 premiers blocs
            // sont réservés au CPU"
            for (int i = 0; i < rom.Length; i++)
                _memory[0x200 + i] = rom[i];
            Reset();
        }

        public void Reset()
        {
            // On vide la pile d'appel
            _CallStack.Clear();
            // On reset PC
            _PC = 0x200;
            // Et l'écran
            Inst_00E0();
        }

        // Les paramètres possibles d'une instruction
        ushort _NNN;
        byte _KK;
        byte _K;
        byte _X;
        byte _Y;

        public MegaChip8()
        {
            _memory[0] = 240; _memory[1] = 144; _memory[2] = 144; _memory[3] = 144; _memory[4] = 240;
            _memory[5] = 32; _memory[6] = 96; _memory[7] = 32; _memory[8] = 32; _memory[9] = 112;
            _memory[10] = 240; _memory[11] = 16; _memory[12] = 240; _memory[13] = 128; _memory[14] = 240;
            _memory[15] = 240; _memory[16] = 16; _memory[17] = 240; _memory[18] = 16; _memory[19] = 240;
            _memory[20] = 144; _memory[21] = 144; _memory[22] = 240; _memory[23] = 16; _memory[24] = 16;
            _memory[25] = 240; _memory[26] = 128; _memory[27] = 240; _memory[28] = 16; _memory[29] = 240;
            _memory[30] = 240; _memory[31] = 128; _memory[32] = 240; _memory[33] = 144; _memory[34] = 240;
            _memory[35] = 240; _memory[36] = 16; _memory[37] = 32; _memory[38] = 64; _memory[39] = 64;
            _memory[40] = 240; _memory[41] = 144; _memory[42] = 240; _memory[43] = 144; _memory[44] = 240;
            _memory[45] = 240; _memory[46] = 144; _memory[47] = 240; _memory[48] = 16; _memory[49] = 240;
            _memory[50] = 240; _memory[51] = 144; _memory[52] = 240; _memory[53] = 144; _memory[54] = 144;
            _memory[55] = 224; _memory[56] = 144; _memory[57] = 224; _memory[58] = 144; _memory[59] = 224;
            _memory[60] = 240; _memory[61] = 128; _memory[62] = 128; _memory[63] = 128; _memory[64] = 240;
            _memory[65] = 224; _memory[66] = 144; _memory[67] = 144; _memory[68] = 144; _memory[69] = 224;
            _memory[70] = 240; _memory[71] = 128; _memory[72] = 240; _memory[73] = 128; _memory[74] = 240;
            _memory[75] = 240; _memory[76] = 128; _memory[77] = 240; _memory[78] = 128; _memory[79] = 128;
        }

        // Emule un cycle du CPU
        public void Emulate()
        {
            // On récupère l'instruction 16bit sur laquelle pointe PC
            // Pour ca on lit 2 byte 8bit que l'on stock sur 16 bits
            ushort opcode = (ushort)((_memory[_PC] << 8) + _memory[_PC + 1]);

            // Puis on incrémente PC d'autant de blocs que l'on vient de lire, soit 2
            _PC += 2;

            // On extrait les différents paramètres possibles d'une instruction à l'avance
            ushort _Code = (ushort)(opcode & 0xF000);
            _NNN = (ushort)(opcode & 0x0FFF);
            _KK = (byte)(opcode & 0x00FF);
            _K = (byte)(opcode & 0x000F);
            _X = (byte)((opcode & 0x0F00) >> 8);
            _Y = (byte)((opcode & 0x00F0) >> 4);

            //System.Console.Out.WriteLine((_PC - 2) + ":" + (_Code >> 12) + "," + _X + "," + _Y + "," + _K + ":" + _I + ":" + _V[0] + ":" + _V[1] + ":" + _V[2] + ":" + _V[3] + ":" + _V[4] + ":" + _V[5] + ":" + _V[6] + ":" + _V[7] + ":" + _V[8] + ":" + _V[9] + ":" + _V[10] + ":" + _V[11] + ":" + _V[12] + ":" + _V[13] + ":" + _V[14] + ":" + _V[15]);

            // Maintenant on décode l'opcode et on le redirige vers la bonne instruction (dispatcher)
            switch (_Code)
            {
                case 0x0000:
                    switch (_KK)
                    {
                        case 0xE0: Inst_00E0(); break;
                        case 0xEE: Inst_00EE(); break;
                    }
                    break;
                case 0x1000: Inst_1NNN(); break;
                case 0x2000: Inst_2NNN(); break;
                case 0x3000: Inst_3XKK(); break;
                case 0x4000: Inst_4XKK(); break;
                case 0x5000: Inst_5XY0(); break;
                case 0x6000: Inst_6XKK(); break;
                case 0x7000: Inst_7XKK(); break;
                case 0x8000:
                    switch (_K)
                    {
                        case 0x0: Inst_8XY0(); break;
                        case 0x1: Inst_8XY1(); break;
                        case 0x2: Inst_8XY2(); break;
                        case 0x3: Inst_8XY3(); break;
                        case 0x4: Inst_8XY4(); break;
                        case 0x5: Inst_8XY5(); break;
                        case 0x6: Inst_8XY6(); break;
                        case 0x7: Inst_8XY7(); break;
                        case 0xE: Inst_8XYE(); break;
                    }
                break;
                case 0x9000: Inst_9XY0(); break;
                case 0xA000: Inst_ANNN(); break;
                case 0xB000: Inst_BNNN(); break;
                case 0xC000: Inst_CXKK(); break;
                case 0xD000: Inst_DXYK(); break;
                case 0xE000:
                    switch (_KK)
                    {
                        case 0x9E: Inst_EX9E(); break;
                        case 0xA1: Inst_EXA1(); break;
                    }
                break;
                case 0xF000:
                switch (_KK)
                    {
                        case 0x07: Inst_FX07(); break;
                        case 0x0A: Inst_FX0A(); break;
                        case 0x15: Inst_FX15(); break;
                        case 0x18: Inst_FX18(); break;
                        case 0x1E: Inst_FX1E(); break;
                        case 0x29: Inst_FX29(); break;
                        case 0x33: Inst_FX33(); break;
                        case 0x55: Inst_FX55(); break;
                        case 0x65: Inst_FX65(); break;
                    }
                break;
            }

            if (_DT > 0)
                _DT--;
            if (_ST > 0)
                _ST--;
        }

        #region Partie graphique

        private byte[] _ScreenBuffer = new byte[8192];

        private void Inst_00E0()
        {
            // noir = {B = 0, G = 0, R = 0, A = 255}
            for (int i = 0; i < _ScreenBuffer.Length; i += 4)
            {
                _ScreenBuffer[i] = 0;
                _ScreenBuffer[i + 1] = 0;
                _ScreenBuffer[i + 2] = 0;
                _ScreenBuffer[i + 3] = 255;
            }
        }

        private void Inst_DXYK()
        {
            // On reset le flag
            _V[0xF] = 0;
            // Pour chaque ligne de pixels à écrire
            for (int i = 0; i < _K; i++)
            {
                byte ligne = _memory[_I + i];
                // Pour chaque pixel dans la ligne
                for (int n = 0; n < 8; n++)
                {
                    // On extrait le pixel à écrire
                    byte pixel = (byte)((byte)(ligne << n) >> 7);
                    // On calcul l'adresse du pixel dans le frame buffer
                    int pixelAddress = _V[_X] + n + (_V[_Y] + i) * 64;
                    pixelAddress *= 4;
                    pixelAddress %= _ScreenBuffer.Length;
                    // On fait la somme des composantes, pour savoir si il y a une couleur
                    int couleur = _ScreenBuffer[pixelAddress] + _ScreenBuffer[pixelAddress + 1] +_ScreenBuffer[pixelAddress + 2];
                    // On l'écrit en XOR
                    lock (_ScreenBuffer)
                    {
                        if ((pixel != 0 && couleur == 0) || (pixel == 0 && couleur != 0))
                        {
                            _ScreenBuffer[pixelAddress] = 255;
                            _ScreenBuffer[pixelAddress + 1] = 255;
                            _ScreenBuffer[pixelAddress + 2] = 255;
                            _ScreenBuffer[pixelAddress + 3] = 255;
                        }
                        else
                        {
                            _ScreenBuffer[pixelAddress] = 0;
                            _ScreenBuffer[pixelAddress + 1] = 0;
                            _ScreenBuffer[pixelAddress + 2] = 0;
                            _ScreenBuffer[pixelAddress + 3] = 255;
                        }
                    }

                    // Et on détecte la collision
                    if (pixel != 0 && couleur != 0)
                        _V[0xF] = 1;
                }
            }
        }

        private void Inst_FX29()
        {
            _I = (ushort)(_V[_X] * 5);
        }

        #endregion

        #region Inputs

        public bool[] Keyboard = new bool[16];

        private void Inst_EX9E()
        {
            if (Keyboard[_V[_X]] == true)
                _PC += 2;
        }

        private void Inst_EXA1()
        {
            if (Keyboard[_V[_X]] == false)
                _PC += 2;
        }

        private void Inst_FX0A()
        {
            while (true)
                for (byte i = 0; i < 16; i++)
                    if (Keyboard[i] == true)
                    {
                        _V[_X] = i;
                        return;
                    }
        }

        #endregion

        #region Misc.

        private void Inst_FX33()
        {
            _memory[_I] = (byte)(_V[_X] / 100);
            _memory[_I + 1] = (byte)((_V[_X] % 100) / 10);
            _memory[_I + 2] = (byte)((_V[_X] % 100) % 10);
        }

        private Random _RandGenerator = new Random();

        private void Inst_CXKK()
        {
            _V[_X] = (byte)(_KK & (byte)_RandGenerator.Next(0, 256));
        }

        private void Inst_7XKK()
        {
            _V[_X] += _KK;
        }

        private void Inst_FX1E() // should set VF if carry
        {
            _I += _V[_X];
        }

        #endregion

        #region Timers

        private void Inst_FX07()
        {
            _V[_X] = _DT;
        }

        private void Inst_FX15()
        {
            _DT = _V[_X];
        }

        private void Inst_FX18()
        {
            _ST = _V[_X];
        }

        #endregion

        #region Math / Bitwise and flags

        private void Inst_8XY4()
        {
            int tmp = _V[_X] + _V[_Y];
            if (tmp > 255)
                _V[0xF] = 1;
            else
                _V[0xF] = 0;
            _V[_X] = (byte)(tmp & 0xFF);
        }

        private void Inst_8XY5()
        {
            if (_V[_X] > _V[_Y])
                _V[0xF] = 1;
            else
                _V[0xF] = 0;
            _V[_X] -= _V[_Y];
        }

        private void Inst_8XY6()
        {
            _V[0xF] = (byte)(_V[_X] & 0x1);
            _V[_X] >>= 1;
        }

        private void Inst_8XY7()
        {
            if (_V[_Y] > _V[_X])
                _V[0xF] = 1;
            else
                _V[0xF] = 0;
            _V[_X] = (byte)(_V[_Y] - _V[_X]);
        }

        private void Inst_8XYE()
        {
            _V[0xF] = (byte)((_V[_X] & 0x80) >> 7);
            _V[_X] <<= 1;
        }

        #endregion

        #region Bitwise

        private void Inst_8XY1()
        {
            _V[_X] |= _V[_Y];
        }

        private void Inst_8XY2()
        {
            _V[_X] &= _V[_Y];
        }

        private void Inst_8XY3()
        {
            _V[_X] ^= _V[_Y];
        }

        #endregion

        #region Read / Write / Load

        private void Inst_6XKK()
        {
            _V[_X] = _KK;
        }

        private void Inst_8XY0()
        {
            _V[_X] = _V[_Y];
        }

        private void Inst_ANNN()
        {
            _I = _NNN;
        }

        private void Inst_FX55()
        {
            for (int i = 0; i <= _X; i++)
                _memory[_I + i] = _V[i];
        }

        private void Inst_FX65()
        {
            for (int i = 0; i <= _X; i++)
                _V[i] = _memory[_I + i];
        }

        #endregion

        #region Conditional branching
        private void Inst_3XKK()
        {
            if (_V[_X] == _KK)
                _PC += 2;
        }

        private void Inst_4XKK()
        {
            if (_V[_X] != _KK)
                _PC += 2;
        }

        private void Inst_5XY0()
        {
            if (_V[_X] == _V[_Y])
                _PC += 2;
        }

        private void Inst_9XY0()
        {
            if (_V[_X] != _V[_Y])
                _PC += 2;
        }
        #endregion

        #region Jump and sub-routine
        private void Inst_1NNN()
        {
            _PC = _NNN;
        }

        // La pile d'appels
        private Stack<ushort> _CallStack = new Stack<ushort>();

        private void Inst_2NNN()
        {
            // On sauvegarde l'adresse de retour
            _CallStack.Push(_PC);
            // On jump à la sous-routine
            _PC = _NNN;
        }

        private void Inst_00EE()
        {
            _PC = _CallStack.Pop();
        }

        private void Inst_BNNN()
        {
            _PC = (ushort)(_NNN + _V[0]);
        }
        #endregion

        public byte[] GetScreenBuffer()
        {
            lock (_ScreenBuffer)
            {
                return _ScreenBuffer;
            }
        }
    }
}
